
#include "USB_CDC.h"


//**********************************************************************************
#define _BSTALL				0x04        //Buffer Stall enable
#define _DTSEN				0x08        //Data Toggle Synch enable
#define _INCDIS				0x10        //Address increment disable

#define _DAT0				0x00        //DATA0 packet expected next
#define _DAT1				0x40        //DATA1 packet expected next
#define _DTSMASK			0x40        //DTS Mask
#define _USIE				0x80        //SIE owns buffer
#define _UCPU				0x00        //CPU owns buffer

#define SETUP_TOKEN			0x0D
#define OUT_TOKEN			0x01
#define IN_TOKEN			0x09

#define CDC_TX_READY		0x00
#define CDC_TX_BUSY			0x01
#define CDC_TX_BUSY_ZLP		0x02
#define CDC_TX_COMPLETING	0x03
//**********************************************************************************

typedef struct
 {
  union
   {
	unsigned char Val;
	struct
	 {
		unsigned BC8 :1;		// bit 8 of the byte count
		unsigned BC9 :1;		// bit 9 of the byte count
		unsigned PID :4;        // Packet Identifier
		unsigned DTS :1;		// Data Toggle Synch Value
		unsigned UOWN:1;		// USB Ownership
	 };
   } STAT;
	unsigned char   CNT;
	unsigned char * ADR;
 } BDT_ENTRY;
 
typedef struct
{
  BDT_ENTRY Out;
  BDT_ENTRY In;
} BDT_ITEM;

typedef union
{
    struct
    {
        unsigned char bmRequestType;		//from table 9-2 of USB2.0 spec
        unsigned char bRequest;				//from table 9-2 of USB2.0 spec
        unsigned int wValue;				//from table 9-2 of USB2.0 spec
        unsigned int wIndex;				//from table 9-2 of USB2.0 spec
        unsigned int wLength;				//from table 9-2 of USB2.0 spec
    };
    struct
    {
        unsigned Recipient:5;				//Device,Interface,Endpoint,Other
        unsigned RequestType:2;				//Standard,Class,Vendor,Reserved
        unsigned DataDir:1;					//Host-to-device,Device-to-host
        unsigned :8;
        unsigned char bFeature;				//DEVICE_REMOTE_WAKEUP,ENDPOINT_HALT
        unsigned :8;
        unsigned :8;
        unsigned :8;
        unsigned :8;
        unsigned :8;
    };
    struct
    {
        unsigned :8;
        unsigned :8;
        unsigned char bDscIndex;			//For Configuration and String DSC Only
        unsigned char bDescriptorType;		//Device,Configuration,String
        unsigned int wLangID;				//Language ID
        unsigned :8;
        unsigned :8;
    }; 
     struct
    {
        unsigned :8;
        unsigned :8;
        unsigned char bDevADR;				//Device Address 0-127
        unsigned char bDevADRH;				//Must equal zero
        unsigned :8;
        unsigned :8;
        unsigned :8;
        unsigned :8;
    };
    struct
    {
        unsigned :8;
        unsigned :8;
        unsigned char bConfigurationValue;	//Configuration Value 0-255
        unsigned char bCfgRSD;				//Must equal zero (Reserved)
        unsigned :8;
        unsigned :8;
        unsigned :8;
        unsigned :8;
    };
   struct
    {
        unsigned :8;
        unsigned :8;
        unsigned char bAltID;				//Alternate Setting Value 0-255
        unsigned char bAltID_H;				//Must equal zero
        unsigned char bIntfID;				//Interface Number Value 0-255
        unsigned char bIntfID_H;			//Must equal zero
        unsigned :8;
        unsigned :8;
    };
    struct
    {
        unsigned :8;
        unsigned :8;
        unsigned :8;
        unsigned :8;
        unsigned EPNum:4;					//Endpoint Number 0-15
        unsigned :3;
        unsigned EPDir:1;					//Endpoint Direction: 0-OUT, 1-IN
        unsigned :8;
        unsigned :8;
        unsigned :8;
    };
} CTRL_TRF_SETUP;

typedef struct
{
 union
  {
	ram unsigned char * bRam;
	rom unsigned char * bRom;
  } pSrc;
 union
  {
	struct
	 {
		unsigned  ctrl_trf_mem	:1;
		unsigned  reserved		:6;
		unsigned  busy			:1;
	 } bits;
	unsigned char Val;
  } info;
 unsigned char bCount;
} IN_PIPE;

//**********************************************************************************

#pragma romdata

static rom unsigned char DeviceDescriptor[] =
{
    0x12,							// Size of this descriptor in bytes
    0x01,							// DEVICE descriptor type
    0x10,0x01,						// USB Spec Release Number in BCD format
    0x02,							// Class Code CDC_DEVICE
    0x00,							// Subclass code
    0x00,							// Protocol code
    0x08,							// Max packet size for EP0
    0xD8,0x04,						// Vendor ID
    0x0A,0x00,						// Product ID: CDC RS-232 Emulation Demo
   (unsigned char)(_DEVICE_ID),		// Device release number in BCD format
   (unsigned char)(_DEVICE_ID >> 8),
    0x00,                   		// Manufacturer string index
    0x00,                   		// Product string index
    0x00,                   		// Device serial number string index
    0x01                    		// Number of possible configurations
};

static rom unsigned char ConfigDescriptor[]={
    0x09,							// Size of this descriptor in bytes
    0x02,							// CONFIGURATION descriptor type
      67,0x00,						// 67 bytes - Total length of data for this cfg 
    0x02,							// Number of interfaces in this cfg
    0x01,							// Index value of this configuration
    0x00,							// Configuration string index
    0xC0,							// Attributes: _DEFAULT | _SELF
     100,							// Max power consumption (100 * 2 = 200 mA)
							
    0x09,							// sizeof(USB_INTF_DSC), Size of this descriptor in bytes
    0x04,							// INTERFACE descriptor type
    0x00,							// Interface Number
    0x00,							// Alternate Setting Number
    0x01,							// Number of endpoints in this intf
    0x02,							// Class code :          COMM_INTF 
    0x02,							// Subclass code:        ABSTRACT_CONTROL_MODEL 
    0x00,							// V25TER Protocol code: Common AT commands ("Hayes(TM)")
    0x00,							// Interface string index

    0x05,							// sizeof(USB_CDC_HEADER_FN_DSC)
    0x24,							// CS_INTERFACE
    0x00,							// DSC_FN_HEADER
    0x10,
	0x01,

    0x04,							// sizeof(USB_CDC_ACM_FN_DSC)
    0x24,							// CS_INTERFACE
    0x02,							// DSC_FN_ACM
    0x02,							// USB_CDC_ACM_FN_DSC_VAL

    0x05,							// sizeof(USB_CDC_UNION_FN_DSC)
    0x24,							// CS_INTERFACE
    0x06,							// DSC_FN_UNION,
    0x00,							// CDC_COMM_INTF_ID
    0x01,							// CDC_DATA_INTF_ID

    0x05,							// sizeof(USB_CDC_CALL_MGT_FN_DSC)
    0x24,							// CS_INTERFACE
    0x01,							// DSC_FN_CALL_MGT
    0x00,
    0x01,							// CDC_DATA_INTF_ID,

    0x07,							// sizeof(USB_EP_DSC)
    0x05,							// Endpoint Descriptor
    0x82,							// _EP02_IN EndpointAddress
    0x03,							// _INTERRUPT Attributes
    0x08,0x00,						// size
    0x02,							// Interval

    0x09,							// Size of INTF descriptor in bytes
    0x04,							// INTERFACE descriptor type
    0x01,							// Interface Number
    0x00,							// Alternate Setting Number
    0x02,							// Number of endpoints in this intf
    0x0A,							// DATA_INTF Class code
    0x00,							// Subclass code
    0x00,							// NO_PROTOCOL Protocol code
    0x00,							// Interface string index
    
    0x07,							// sizeof(USB_EP_DSC)
    0x05,							// Endpoint Descriptor
    0x03,							// _EP03_OUT EndpointAddress
    0x02,							// _BULK Attributes
      64,0x00,						// size: 64 bytes
    0x00,      						// Interval

    0x07,							// sizeof(USB_EP_DSC)
    0x05,							// Endpoint Descriptor
    0x83,							// _EP03_IN EndpointAddress
    0x02,							// _BULK Attributes
      64,0x00,						// size: 64 bytes
    0x00,							// Interval
};

//**********************************************************************************

#if defined(__18F14K50) || defined(__18F13K50) || defined(__18LF14K50) || defined(__18LF13K50)
	#pragma udata USB_BDT=0x200     //See Linker Script,usb2:0x200-0x2FF(256-byte)
#else
	#pragma udata USB_BDT=0x400     //See Linker Script,usb4:0x400-0x4FF(256-byte)
#endif

static BDT_ITEM				BDT[4];
static CTRL_TRF_SETUP		SetupPkt;
static unsigned char 		CtrlTrfData[8];

unsigned char				USB_CDC_RC_BUFF[64];
unsigned char				USB_CDC_TX_BUFF[64];

//**********************************************************************************

#pragma udata

static unsigned char		cdc_rx_len;		// total rx length
static unsigned char		cdc_trf_state;	// States are defined cdc.h
static unsigned char		cdc_tx_len;		// total tx length
static unsigned char *		cdc_tx_adr;

static unsigned char		d_en_cmd_resp[8];

static unsigned char		USBActiveConfiguration;
static unsigned char		USBAlternateInterface;

static unsigned char		shortPacketStatus;
static unsigned char		controlTransferState;

static unsigned char		notice_change;

static IN_PIPE				inPipes;

static unsigned char		RemoteWakeUp;

static struct
 {
	union
	 {
		ram unsigned char * Ram;
		BDT_ENTRY * 		Bdt;
		rom unsigned char * Rom;
	 } Src;
	ram unsigned char * Dst;
	ram unsigned char   Cnt;
 } CopyRef;

CONTROL_SIGNAL				USB_CDC_DTE_RTS;
unsigned char				USB_CDC_STATE;
LINE_CODING					USB_CDC_LINE_CODING;

//**********************************************************************************

#pragma code

//**********************************************************************************

static void RAM_COPY(void)
{
	while(CopyRef.Cnt--) *CopyRef.Dst++ = *CopyRef.Src.Ram++;
}

//**********************************************************************************

static void USBCheckCDCRequest(void)
{
	if(SetupPkt.Recipient != 0x01) return; // != RCPT_INTF

	if(SetupPkt.RequestType != 1) return;// != CLASS

	if((SetupPkt.bIntfID) && (SetupPkt.bIntfID != 0x01)) return; // != CDC_DATA_INTF_ID

	switch(SetupPkt.bRequest)
	 {
        case 0x00: //SEND_ENCAPSULATED_COMMAND:
            inPipes.pSrc.bRam              = & d_en_cmd_resp[0];
            inPipes.bCount                 = 8; // dummy_length
            inPipes.info.bits.ctrl_trf_mem = 1; // USB_INPIPES_RAM;
            inPipes.info.bits.busy         = 1;
            return;

        case 0x01: // GET_ENCAPSULATED_RESPONSE:
            inPipes.pSrc.bRam      = & d_en_cmd_resp[0];
            inPipes.info.bits.busy = 1;
            return;

        case 0x20: // SET_LINE_CODING:
			notice_change = 1;
            return;
            
        case 0x21: // GET_LINE_CODING:
			inPipes.pSrc.bRam = (unsigned char *) & USB_CDC_LINE_CODING;
			inPipes.bCount    = 7;
			inPipes.info.Val  = 0x01 | 0x80; // USB_INPIPES_RAM | USB_INPIPES_BUSY
			return;

        case 0x22: //SET_CONTROL_LINE_STATE:
			*((unsigned char *) & USB_CDC_DTE_RTS) = (unsigned char) SetupPkt.wValue;
			USB_CDC_DTE_RTS.Change = 1;
            inPipes.info.bits.busy = 1;
	 };
}

//**********************************************************************************

static void USBTransferOnePacketIn(void)
{
	BDT[3].In.STAT.Val ^= _DTSMASK;
	BDT[3].In.STAT.Val &= _DTSMASK;
	BDT[3].In.STAT.Val |= _USIE | _DTSEN;
}

static void USBTransferOnePacketOut(void)
{
	BDT[3].Out.STAT.Val ^= _DTSMASK;
	BDT[3].Out.ADR       = & USB_CDC_RC_BUFF[0];
	BDT[3].Out.CNT       = 64;
	BDT[3].Out.STAT.Val &= _DTSMASK;
	BDT[3].Out.STAT.Val |= _USIE | _DTSEN;
}

//**********************************************************************************

static void USBDeviceInit(void)
{
	UEIR  = 0x00;  
	UIR   = 0x00; 
	UEIE  = 0x9F;
	UADDR = 0x00;
	UEP1  = 0x00;
	UEP2  = 0x00;
    UIE   = 0xFB; 

    UCONbits.PPBRST = 1;

	inPipes.info.Val           = 0;
	notice_change              = 0;
	USBActiveConfiguration     = 0;     
	USB_CDC_STATE              = DETACHED_STATE;

    UCONbits.PPBRST   = 0;

	UEP0 = 0x06 | 0x10; // EP_CTRL | USB_HANDSHAKE_ENABLED;

	while(UIRbits.TRNIF)
	 {
		UIRbits.TRNIF = 0;
		ClrWdt();
	 };

	UCONbits.PKTDIS   = 0;
}

//**********************************************************************************

static void USBStdGetDscHandler(void)
{
	if(SetupPkt.bmRequestType != 0x80) return;

	inPipes.info.Val = 0x00 | 0x80; // USB_INPIPES_ROM | USB_INPIPES_BUSY

	switch(SetupPkt.bDescriptorType)
	 {
		case 0x01: // USB_DESCRIPTOR_DEVICE:
			inPipes.pSrc.bRom =  & DeviceDescriptor[0];
			inPipes.bCount    = sizeof(DeviceDescriptor);
			return;

		case 0x02: // USB_DESCRIPTOR_CONFIGURATION:
			inPipes.pSrc.bRom = & ConfigDescriptor[0];
			inPipes.bCount    = 67;
			return;

		case 0x03: // USB_DESCRIPTOR_STRING:
			inPipes.info.Val = 0;
	 };
}

//**********************************************************************************

static void USBStdSetCfgHandler(void)
{
	inPipes.info.bits.busy = 1;

	UEP1 = 0x00;
	UEP2 = 0x00;

	USBAlternateInterface = 0x00;

	USBActiveConfiguration = SetupPkt.bConfigurationValue;

	if(SetupPkt.bConfigurationValue == 0)
	 {
		USB_CDC_STATE = ADDRESS_STATE;
		return;
	 };

	USB_CDC_STATE = CONFIGURED_STATE;

	USB_CDC_LINE_CODING.BaudRate     = 19200;     // baud rate
	USB_CDC_LINE_CODING.CharFormat   =  0x00;     // 1 stop bit
	USB_CDC_LINE_CODING.ParityType   =  0x00;     // None
	USB_CDC_LINE_CODING.DataBits     =  0x08;     // 5,6,7,8, or 16
	USB_CDC_LINE_CODING.Change       =     1;

	*((unsigned char *) & USB_CDC_DTE_RTS) =  0;
	USB_CDC_DTE_RTS.Change                 =  1;

    cdc_trf_state           = CDC_TX_READY;
    cdc_rx_len              = 0;

	UEP2                    = 0x02 |0x10 | 0x08; 		// USB_IN_ENABLED |USB_HANDSHAKE_ENABLED | USB_DISALLOW_SETUP	
	BDT[2].In.STAT.DTS      = 1;
	BDT[2].In.STAT.UOWN     = 0;

	UEP3                    = 0x02 |0x04 | 0x10 | 0x08; // USB_IN_ENABLED |USB_OUT_ENABLED | USB_HANDSHAKE_ENABLED | USB_DISALLOW_SETUP
	BDT[3].In.STAT.DTS      = 1;
	BDT[3].In.STAT.UOWN     = 0;
	BDT[3].Out.STAT.DTS     = 1;
	BDT[3].Out.STAT.UOWN    = 0;
	USBTransferOnePacketOut();
}

//**********************************************************************************

static void Set_BDT_SetupPkt_EPNum(void)
{
	CopyRef.Src.Bdt = & BDT[SetupPkt.EPNum].Out;
	if(SetupPkt.EPDir) CopyRef.Src.Bdt++;
}

//**********************************************************************************

static void USBStdGetStatusHandler(void)
{
	CtrlTrfData[0] = 0;                 // Initialize content
	CtrlTrfData[1] = 0;

	switch(SetupPkt.Recipient)
	 {
		case 0x00: // RCPT_DEV:
			inPipes.info.bits.busy = 1;
            CtrlTrfData[0]|=0x01;
			if(RemoteWakeUp) CtrlTrfData[0]|=0x02;
			break;

		case 0x01: // RCPT_INTF:
			inPipes.info.bits.busy = 1;     // No data to update
			break;

		case 0x02: // RCPT_EP:
			inPipes.info.bits.busy = 1;
			Set_BDT_SetupPkt_EPNum();
			if(CopyRef.Src.Bdt->STAT.Val & _BSTALL) CtrlTrfData[0] = 0x01;    // Set bit0
	 };

    if(inPipes.info.bits.busy == 1)
	 {
        inPipes.pSrc.bRam              = & CtrlTrfData[0];
        inPipes.info.bits.ctrl_trf_mem = 1; // USB_INPIPES_RAM;
		inPipes.bCount                 = 2;
	 };
}

//**********************************************************************************

static void USBStdFeatureReqHandler(void)
{            
	if((SetupPkt.bFeature == 1) && (SetupPkt.Recipient == 0))
	 {
		inPipes.info.bits.busy = 1;
		RemoteWakeUp = 0;
        if(SetupPkt.bRequest == 3) RemoteWakeUp = 1; // SET_FEATURE
	 };

	if((SetupPkt.bFeature == 0) && (SetupPkt.Recipient == 2) && SetupPkt.EPNum)
	 {
		inPipes.info.bits.busy = 1;
		Set_BDT_SetupPkt_EPNum();

		CopyRef.Src.Bdt->STAT.Val = _USIE | _BSTALL;
		if(SetupPkt.bRequest != 3)
		 {
			*((unsigned int *) (& UEP0 + SetupPkt.EPNum)) &= 0xFE; // ~UEP_STALL
			CopyRef.Src.Bdt->STAT.Val = _USIE | _DAT1 | _DTSEN;
			if(SetupPkt.EPDir) CopyRef.Src.Bdt->STAT.Val = _UCPU | _DAT1;   
		 };
    };
}

//**********************************************************************************

static void USBCheckStdRequest(void)
{
	if(SetupPkt.RequestType) return; // ! STANDARD

	switch(SetupPkt.bRequest)
	 {
		case 0x05: // SET_ADR:
			inPipes.info.bits.busy = 1;					// This will generate a zero length packet
			USB_CDC_STATE          = ADR_PENDING_STATE;	// Update state only
			return;

		case 0x06: // GET_DSC:
			USBStdGetDscHandler();
			return;

		case 0x09: // SET_CFG:
			USBStdSetCfgHandler();
			return;

		case 0x08: // GET_CFG:
			inPipes.pSrc.bRam              = (unsigned char*) & USBActiveConfiguration;	// Set Source
			inPipes.info.bits.ctrl_trf_mem = 1; 										// USB_INPIPES_RAM;
			inPipes.bCount                 = 1;											// Set data count
			inPipes.info.bits.busy         = 1;
			return;

		case 0x00: // GET_STATUS:
			USBStdGetStatusHandler();
			return;

		case 0x01: // CLR_FEATURE:
		case 0x03: // SET_FEATURE:
			USBStdFeatureReqHandler();
			return;

		case 0x0A: // GET_INTF:
			inPipes.pSrc.bRam              = (unsigned char *) & USBAlternateInterface + SetupPkt.bIntfID;	// Set source
			inPipes.info.bits.ctrl_trf_mem = 1;																// USB_INPIPES_RAM;
			inPipes.bCount                 = 1;																 // Set data count
			inPipes.info.bits.busy         = 1;
			return;

		case 0x0B: // SET_INTF:
			inPipes.info.bits.busy = 1;
			USBAlternateInterface = SetupPkt.bAltID;
	 };
}

//**********************************************************************************

static void USBCtrlTrfTxService(void)
{
	CopyRef.Cnt = 8;

    if(inPipes.bCount < 8)
	 {
        CopyRef.Cnt = inPipes.bCount;
		if(shortPacketStatus == 1) shortPacketStatus = 2; // SHORT_PKT_SENT;
        if(shortPacketStatus == 0) shortPacketStatus = 1; // SHORT_PKT_PENDING;
	 };

	BDT[0].In.STAT.BC9 = 0;
	BDT[0].In.STAT.BC8 = 0;
	BDT[0].In.CNT      = CopyRef.Cnt;

    inPipes.bCount = inPipes.bCount - CopyRef.Cnt;

    CopyRef.Dst    = & CtrlTrfData[0];

	if(inPipes.info.bits.ctrl_trf_mem == 0) // USB_INPIPES_ROM
	 {
        while(CopyRef.Cnt--) *CopyRef.Dst++ = *inPipes.pSrc.bRom++;
		return;
	 };

	CopyRef.Src.Ram = inPipes.pSrc.bRam;
	RAM_COPY(); inPipes.pSrc.bRam  = CopyRef.Src.Ram;
}

//**********************************************************************************

static void USBCtrlEPServiceComplete(void)
{
	UCONbits.PKTDIS = 0;
	BDT[0].Out.CNT  = 8; // USB_EP0_BUFF_SIZE;

	if(inPipes.info.bits.busy == 0)
	 {
		if(notice_change)
		 {
			controlTransferState = 2; // CTRL_TRF_RX;

			BDT[0].In.CNT        = 0;
            BDT[0].In.STAT.Val   = _USIE | _DAT1 | _DTSEN;
			BDT[0].Out.ADR       = (unsigned char *) & CtrlTrfData;
			BDT[0].Out.STAT.Val   = _USIE|_DAT1|_DTSEN;
			return;
		 };

		BDT[0].Out.ADR      = (unsigned char*) & SetupPkt;
		BDT[0].Out.STAT.Val = _USIE | _DAT0 | _DTSEN | _BSTALL;
		BDT[0].In.STAT.Val  = _USIE | _BSTALL; 
		return;
     };

	if(notice_change) return;

	if(SetupPkt.DataDir) //DEV_TO_HOST
	 {
		if(SetupPkt.wLength < (unsigned int) inPipes.bCount) inPipes.bCount = (unsigned char) SetupPkt.wLength;
		USBCtrlTrfTxService();
		controlTransferState = 1;
		BDT[0].Out.ADR       = (unsigned char *) & SetupPkt;
		BDT[0].Out.STAT.Val  = _USIE;
		BDT[0].Out.CNT       = 8;
		BDT[0].Out.ADR       = (unsigned char *)&SetupPkt;
		BDT[0].Out.STAT.Val  = _USIE;
		BDT[0].In.ADR        = (unsigned char *) & CtrlTrfData;
		BDT[0].In.STAT.Val   = _USIE|_DAT1|_DTSEN;
		return;
	 };

	controlTransferState     = 2; // CTRL_TRF_RX;
	BDT[0].In.CNT            = 0;
	BDT[0].In.STAT.Val       = _USIE|_DAT1|_DTSEN;
	BDT[0].Out.ADR           = (unsigned char*) & CtrlTrfData;
	BDT[0].Out.STAT.Val      = _USIE|_DAT1|_DTSEN;
}

//**********************************************************************************

static void USBCtrlEPService(void)
{
	if(USTAT) goto USBCtrlTrfInHandler;
	 
	if(BDT[0].Out.STAT.PID == SETUP_TOKEN)
	 {
		// USBCtrlTrfSetupHandler

		if(BDT[0].In.STAT.UOWN) BDT[0].In.STAT.Val = _UCPU;  

		shortPacketStatus    = 0; //SHORT_PKT_NOT_USED;
		controlTransferState = 0; // WAIT_SETUP;
		inPipes.bCount       = 0;
		inPipes.info.Val     = 0;

		USBCheckStdRequest();


		USBCheckCDCRequest();
		USBCtrlEPServiceComplete();

		return;
	 };

	//USBCtrlTrfOutHandler();
	if(controlTransferState == 2)
	 {
		//USBCtrlTrfRxService();
		CopyRef.Cnt                =   7;
		CopyRef.Src.Ram            = & CtrlTrfData[0];
		CopyRef.Dst                = (unsigned char *) & USB_CDC_LINE_CODING;
		RAM_COPY();
		USB_CDC_LINE_CODING.Change = 1;
		BDT[0].Out.ADR             = (unsigned char *) & SetupPkt;
		return;
	 };

	goto USBPrepareForNextSetupTrf;

USBCtrlTrfInHandler:

	if(USTAT != 0x04) return; // != USTAT_EP0_IN

	if(USB_CDC_STATE == ADR_PENDING_STATE)
	 {
        UADDR         = SetupPkt.bDevADR;
		USB_CDC_STATE = DEFAULT_STATE;
		if(UADDR) USB_CDC_STATE = ADDRESS_STATE;
	 };

	if(controlTransferState == 1) // == CTRL_TRF_TX
	 {
		BDT[0].In.ADR = (unsigned char *) CtrlTrfData;
		USBCtrlTrfTxService();

		if(shortPacketStatus == 2) // SHORT_PKT_SENT
		 {
            BDT[0].In.STAT.Val = _USIE | _BSTALL;
			return;
		 };

		if(BDT[0].In.STAT.DTS == 0)
		 {
			BDT[0].In.STAT.Val = _USIE|_DAT1|_DTSEN;
			return;
		 };

		BDT[0].In.STAT.Val = _USIE | _DAT0 | _DTSEN;
		return;
	 };

USBPrepareForNextSetupTrf:

	notice_change = 0;

	if(controlTransferState != 2)						goto wait_setup;
	if(UCONbits.PKTDIS      == 0)						goto wait_setup;
	if(BDT[0].Out.CNT       != sizeof(CTRL_TRF_SETUP))	goto wait_setup;
	if(BDT[0].Out.STAT.PID  != SETUP_TOKEN)				goto wait_setup;
	if(BDT[0].Out.STAT.UOWN)							goto wait_setup;

	BDT[0].Out.ADR  = (unsigned char *) & SetupPkt;
	CopyRef.Src.Ram = & CtrlTrfData[0];
	CopyRef.Dst     = (unsigned char *) & SetupPkt;
	CopyRef.Cnt     = sizeof(CTRL_TRF_SETUP);
	RAM_COPY();
	return;

wait_setup:

	controlTransferState = 0; // WAIT_SETUP;
	BDT[0].Out.CNT       = 8; //USB_EP0_BUFF_SIZE;
	BDT[0].Out.ADR       = (unsigned char*) & SetupPkt;
	BDT[0].Out.STAT.Val  = _USIE | _DAT0 | _DTSEN | _BSTALL;
	BDT[0].In.STAT.Val   = _UCPU; 

}

//**********************************************************************************

static void USBDeviceTasks(void)
{
	if(USB_CDC_STATE == DETACHED_STATE)
	 {
		UCON = 0;
        UIE  = 0;

		while(UCONbits.USBEN == 0)
		 {
			UCONbits.USBEN = 1;
			ClrWdt();
		 };

		USB_CDC_STATE = ATTACHED_STATE;
        UCFG = 0X10 | 0x00 | 0x04; // USB_PULLUP_ENABLE | USB_INTERNAL_TRANSCEIVER | USB_FULL_SPEED;
		while(UCONbits.SE0);
		UIR = 0;
        UIE = 0;
        UIEbits.URSTIE = 1;
        UIEbits.IDLEIE = 1;
        USB_CDC_STATE  = POWERED_STATE;
		PIE2bits.USBIE = 1;
	 };

	if(USB_CDC_STATE == ATTACHED_STATE)
	 {
		if(UCONbits.SE0 == 0)
		 {
			UIR            = 0;				// Clear all USB interrupts
			UIE            = 0;				// Mask all USB interrupts
			UIEbits.URSTIE = 1;				// Unmask RESET interrupt
			UIEbits.IDLEIE = 1;				// Unmask IDLE interrupt
			USB_CDC_STATE  = POWERED_STATE;
		 };
	 };

	if(UIRbits.ACTVIF && UIEbits.ACTVIE)	// USBWakeFromSuspend
	 {
		UCONbits.SUSPND = 0;
		UIEbits.ACTVIE  = 0;
		while(UIRbits.ACTVIF) UIRbits.ACTVIF = 0;
	 };

	if(UCONbits.SUSPND) return;

    if(UIRbits.URSTIF && UIEbits.URSTIE)
     {
        USBDeviceInit();
        USB_CDC_STATE        = DEFAULT_STATE;
        BDT[0].Out.ADR       = (unsigned char*) & SetupPkt;
        BDT[0].Out.CNT       = 8;
        BDT[0].Out.STAT.Val  = _USIE | _DAT0 | _DTSEN | _BSTALL;
		UIRbits.URSTIF       = 0;
     };

	if(UIRbits.IDLEIF && UIEbits.IDLEIE)	// USBSuspend
	 { 
		UIEbits.ACTVIE  = 1;
		UIRbits.IDLEIF  = 0;
		UCONbits.SUSPND = 1;
		UIRbits.IDLEIF  = 0;
	 };

    if(UIRbits.SOFIF && UIEbits.SOFIE)
     {
        UIRbits.SOFIF = 0;
     };

    if(UIRbits.STALLIF && UIEbits.STALLIE)
	 {
		if(UEP0bits.EPSTALL == 1)
		 {
			if((BDT[0].Out.STAT.Val == _USIE) && (BDT[0].In.STAT.Val == (_USIE | _BSTALL)))
				BDT[0].Out.STAT.Val = _USIE | _DAT0 | _DTSEN | _BSTALL;
			UEP0bits.EPSTALL = 0;   
		 };

		UIRbits.STALLIF = 0;
	 };

//    if(UIRbits.UERRIF && UIRbits.UERRIF)
//    {
        UEIR = 0;               // This clears UERRIF
//    };

    if(USB_CDC_STATE < DEFAULT_STATE) return;

	if(UIEbits.TRNIE && UIRbits.TRNIF)
	 {
		while(UIRbits.TRNIF)
		 {
			USBCtrlEPService();
			UIRbits.TRNIF = 0;
		 };
	 };

//  CDCTxService

	if(BDT[3].In.STAT.UOWN) return;

	if(cdc_trf_state == CDC_TX_COMPLETING) cdc_trf_state = CDC_TX_READY;
        
	if(cdc_trf_state == CDC_TX_READY) return;
    
    if(cdc_trf_state == CDC_TX_BUSY_ZLP)
    {
		BDT[3].In.ADR       = 0;
		BDT[3].In.CNT       = 0;
		USBTransferOnePacketIn();
        cdc_trf_state = CDC_TX_COMPLETING;
    };

	if(cdc_trf_state != CDC_TX_BUSY) return;

	CopyRef.Cnt   = cdc_tx_len;
	if(cdc_tx_len > 64) CopyRef.Cnt = 64;

	cdc_tx_len -= CopyRef.Cnt;

	if(cdc_tx_len == 0)
	 {
		cdc_trf_state = CDC_TX_COMPLETING;
		if(CopyRef.Cnt == 64) cdc_trf_state = CDC_TX_BUSY_ZLP;
	 };

	BDT[3].In.CNT   = CopyRef.Cnt; 
 	BDT[3].In.ADR   = & USB_CDC_TX_BUFF[0];  	  
	CopyRef.Dst     = & USB_CDC_TX_BUFF[0];
	CopyRef.Src.Ram =   cdc_tx_adr;
	RAM_COPY(); cdc_tx_adr  =   CopyRef.Src.Ram;
	
	USBTransferOnePacketIn();
}

//**********************************************************************************
//**********************************************************************************
//**********************************************************************************
//**********************************************************************************
//**********************************************************************************

void USB_CDC_OPEN(void)
{
	USB_CDC_LINE_CODING.Change = 0;
	USB_CDC_DTE_RTS.Change     = 0;

#if defined(_INTERRUPT_HIGH_PRIORITY)
	IPR2bits.USBIP = 1;		//	 
 #elif defined(_INTERRUPT_LOW_PRIORITY)
	IPR2bits.USBIP = 0;		//	 
 #else
    #error Invalid interrupt priority!
#endif

	USBDeviceInit();

	USBDeviceTasks();
}

//**********************************************************************************

void USB_CDC_CLOSE(void)
{
	PIE2bits.USBIE = 0;
	UCON           = 0;
}

//**********************************************************************************

void USB_CDC_INTERRUPT(void)
{
	if(PIE2bits.USBIE && PIR2bits.USBIF)
	 {
		PIR2bits.USBIF = 0;
		USBDeviceTasks();
	 };
}

//**********************************************************************************

unsigned char USB_CDC_RC_COUNT(void)
{
	if(PIE2bits.USBIE == 0) return 0;
	if(USB_CDC_STATE < CONFIGURED_STATE) return 0;
	if(BDT[3].Out.STAT.UOWN) return 0;
	return BDT[3].Out.CNT;
}

//**********************************************************************************

unsigned char USB_CDC_TX_BUSY(void)
{
	if(PIE2bits.USBIE == 0) return 1;
	return cdc_trf_state;
}

//**********************************************************************************

void WRITE_TO_USB_CDC(void * Buff, unsigned char Count)
{
	while(USB_CDC_TX_BUSY());

	cdc_tx_adr    = Buff;
	cdc_tx_len    = Count;
	cdc_trf_state = CDC_TX_BUSY;
}

//**********************************************************************************

unsigned char READ_FROM_USB_CDC(void * Buff)
{
 unsigned char Count;

	Count = USB_CDC_RC_COUNT();

	if(Count)
	 {
		CopyRef.Src.Ram   = & USB_CDC_RC_BUFF[0];
		CopyRef.Dst       = Buff;
		CopyRef.Cnt       = Count;
		RAM_COPY();
		PIE2bits.USBIE = 0;
		USBTransferOnePacketOut();
		PIE2bits.USBIE = 1;
	 };

	return Count;
}

//**********************************************************************************
